#!/usr/bin/env python

import argparse

import csv
import dendropy

def getOpts():

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="""

            Root groups at ggpy results

	* Example:

	    $ root_groups.py [ggpy results] -r [file with outgroups]

            note: ggpy results are files containing the 
                  following format:

                   alignment	tree_id	 group	rank	au_test
                   ...         ...      ...    ...     ...
""",
                                     epilog="")
    parser.add_argument('filenames',
                        nargs  = "*",
                        help='Filenames')
    parser.add_argument('-r','--root_group',
                        metavar="",
                        type= str,
                        required=True,
                        help='File with list of outgroups')
    parser.add_argument('-s', '--suffix',
                        metavar = "",
                        type    = str,
                        default = "_rooted_groups.tsv",
                        help    = '[Optional] suffix [Default = _rooted_groups.tsv]')                    
    args = parser.parse_args()
    return args

def main():
    args = getOpts()
    outgroups = []
    with open(args.root_group, 'r') as f:
        for i in f.readlines():
            outgroups.append(i.strip())

    file_myrows = {}

    for filename in args.filenames:
        myrows = []
        with open(filename, 'r') as f:
            reader = csv.reader(f, delimiter = "\t")
            for row in reader:
                myrows.append( row )

        file_myrows[filename] = myrows

    
    transformed = {}

    for file,myrows in file_myrows.items():

        new_rows = []
        for row in myrows:

            tree_str = row[2]

            if tree_str == 'group':
                continue

            if transformed.__contains__(tree_str):
                row[2] = transformed[tree_str]
                new_rows.append(row)
                continue
            
            from_tree = dendropy.Tree.get_from_string( 
                                        src     = tree_str, 
                                        schema  = 'newick',
                                        rooting = 'default-rooted',
                                        preserve_underscores = True 
                                    )

            root_mrca = from_tree.mrca(taxon_labels = outgroups)
            from_tree.reroot_at_edge(root_mrca.edge)
            to_tree_str = (from_tree
                                .as_string(schema = 'newick')
                                .strip()
                                .replace("[&R] ", "")
                                .replace("'", "")
                            )
                            
            row[2] = to_tree_str

            new_rows.append(row)
            transformed[tree_str] = to_tree_str

        ggi_rooted = [['alignment', 'tree_id', 'group', 'rank', 'au_test']] + new_rows

        with open(file + args.suffix, 'w') as f:
            writer = csv.writer(f, delimiter = "\t")
            writer.writerows(ggi_rooted)
    
if __name__ == "__main__":
    main()





import csv
import sys
import collections



import shap
import joblib
import xgboost
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn.metrics import  accuracy_score, plot_confusion_matrix

# """
# 1,2
# 1,3
# 2,3
# """

class Post_ggi:

    def __init__(self, 
                 feature_file = None,
                 all_ggi_results = None,
                 file_comparisons = None,
                 model_prefix = "post_ggi",
                 cnfx_ncols = 3,
                 threads = 1
                 ) -> None:

        self.feature_file = feature_file
        self.all_ggi_results = all_ggi_results
        self.file_comparisons = file_comparisons

        self.model_prefix = model_prefix
        self.cnf_ncols = cnfx_ncols
        self.threads = threads

        self.drop_columns = [
            'Group', 'aln_base', 
            'SymPval', 'MarPval', 
            'IntPval'
        ]

        # previously tuned hyperparameters
        self.gamma = 0.14210526315789473
        self.learning_rate = 0.03
        self.max_depth = 4
        self.reg_lambda = 0.018
        self.n_estimators = 4200

    @property    
    def ggi_df(self):

        if not self.all_ggi_results:
            return None

        ggi_rows = []
        with open(self.all_ggi_results, 'r') as f:
            reader = csv.reader(f, delimiter = '\t')
            for row in reader:
                ggi_rows.append(row)

        return ggi_rows

    @property
    def tree_id_comp(self):

        if not self.file_comparisons:
            return None

        tree_id_comp = []
        with open(self.file_comparisons, 'r') as f:
            reader = csv.reader(f, delimiter = ',')
            for row in reader:
                tree_id_comp.append(row)

        return tree_id_comp

    @property
    def features(self):
        return pd.read_csv(self.feature_file, sep = '\t')

    def subset_tree_id(self, tree_id_comp1):
        """
        subset tree_id_comp df
        by a tree_id
        """
        ha,hb = tree_id_comp1

        if not self.ggi_df:
            return None

        subset_df = []
        for row in self.ggi_df:

            rank = row[3]
            if rank != '1':
                continue

            aln_base = row[0]
            tree_id  = row[1]

            # contained = set(tree_id_comp1) & set([tree_id])

            if tree_id == ha or tree_id == hb:
                subset_df.append([aln_base, tree_id])

        if not subset_df:
            sys.stderr.write("Tree ids '%s' do not match with\n" % ", ".join(tree_id_comp1))
            sys.stderr.write("'%s' data frame\n" % self.all_ggi_results)
            sys.stderr.flush()
            return None

        return pd.DataFrame(subset_df, columns = ['alignment', 'hypothesis'])
        
    def make_specific_prefix(self, tree_id_comp1):
        return "%s_h%s_h%s" % (self.model_prefix, tree_id_comp1[0], tree_id_comp1[1])

    def shap_plots(self, xgb_clf_no_nor, all_num, new_prefix, _groups_dict):

        # xgb_clf_no_nor, all_num, new_prefix = xgb_clf_no_nor, all_num, new_prefix

        bee_20_filename = "20best_beeswarm_%s.png" % new_prefix
        bee_all_filename = "all_beeswarm_%s.png" % new_prefix

        explainer   = shap.Explainer(xgb_clf_no_nor, all_num)
        shap_values = explainer(all_num)

        shap.plots.beeswarm(shap_values,
                            max_display = 20,
                            show = False)

        plt.title(_groups_dict[True], loc = 'right')
        plt.title(_groups_dict[False], loc = 'left')
        plt.tight_layout(pad=0.05)
        plt.savefig(bee_20_filename, dpi = 330)
        plt.close()

        # explainer   = shap.Explainer(xgb_clf_no_nor, all_num)
        shap_values = explainer(all_num)
        shap.plots.beeswarm(shap_values,
                            max_display = len(all_num.columns),
                            show = False)
        plt.title(_groups_dict[True], loc = 'right')
        plt.title(_groups_dict[False], loc = 'left')
        plt.tight_layout(pad=0.05)
        plt.savefig(bee_all_filename, dpi = 330)
        plt.close()

    def create_metdata(self, new_prefix, values):

        metadata_filename = '%s_metadata.tsv' % new_prefix
        columns = [
            'model_filename',
            'accuracy',
            'pos_hypo',
            'neg_hypo'
        ]

        df = list(zip(columns, values))

        with open(metadata_filename, 'w') as f:
            writer = csv.writer(f, delimiter = "\t")
            writer.writerows(df)

        sys.stdout.write('\tmetadata written at: "%s"\n\n' % metadata_filename)
        sys.stdout.flush()
        
    def xgboost_classifier(self, tree_id_comp1):

        # tree_id_comp1 = self.tree_id_comp[0]

        if not tree_id_comp1:
            return None

        dataset = "H%s-H%s" % tuple(tree_id_comp1)
        new_prefix = self.make_specific_prefix(tree_id_comp1)
        model_filename = '%s.sav' % new_prefix

        sys.stdout.write('Processing: "%s" dataset\n' % dataset)
        sys.stdout.flush()

        features = self.features
        target = self.subset_tree_id(tree_id_comp1)

        # merge 
        aln_base,hypothesis = target.columns
        aln_feature = features.columns[0]

        target = target.rename({aln_base: aln_feature}, axis = 1)
        merged_dataset = features.merge(target, on = aln_feature, how='left')
        new_df = merged_dataset[merged_dataset[hypothesis].notna()].reset_index(drop=True)

        for c in self.drop_columns:
            try:
                new_df = new_df.drop( c, 1 )
            except KeyError:
                pass

        # hypotheses definition
        _groups_dict = { 
            True  : 'H%s' % tree_id_comp1[0],
            False : 'H%s' % tree_id_comp1[1]
        }

        split = StratifiedShuffleSplit(n_splits = 1, test_size = 0.25, random_state = 42)
        for train_index, _ in split.split(new_df, new_df[hypothesis]):
            strat_train_set = new_df.loc[train_index]
            # strat_test_set  = new_df.loc[test_index]
        
        train_num = strat_train_set.drop(hypothesis, axis=1)
        train_labels = strat_train_set[hypothesis] == tree_id_comp1[0]
    
        all_num  = new_df.drop(hypothesis, axis=1)
        all_labels = new_df[hypothesis] == tree_id_comp1[0]

        fr_tr = collections.Counter(all_labels)

        if len(fr_tr) <= 1:
            sys.stdout.write("Less than two labels found in the '%s' dataset\n" % dataset)
            sys.stdout.flush()
            return None
        
        scale_pos_weight = fr_tr[False]/fr_tr[True]

        xgb_clf_no_nor = xgboost.XGBClassifier(
                            objective = 'binary:logistic',
                            colsample_bytree = 0.5,
                            subsample = 0.9,

                            use_label_encoder = not True,
                            scale_pos_weight = scale_pos_weight,
                            
                            gamma = self.gamma,
                            learning_rate = self.learning_rate,
                            max_depth = self.max_depth,
                            reg_lambda = self.reg_lambda,
                            n_estimators = self.n_estimators,

                            n_jobs = self.threads
                        )

        sys.stdout.write("\tRunning the XGBoost classifier\r")
        sys.stdout.flush()

        xgb_clf_no_nor.fit(all_num, all_labels.tolist(),
                           eval_set = [ (train_num, train_labels.tolist()) ],
                           early_stopping_rounds = 2000,
                           verbose = False,
                           eval_metric = 'aucpr'
                        )

        accuracy = accuracy_score(all_labels, xgb_clf_no_nor.predict(all_num))

        sys.stdout.write("\tRunning the XGBoost classifier, overall accuracy: %s\n" % round(accuracy, 6))
        sys.stdout.flush()

        sys.stdout.write("\tCalculating SHAP values\n")
        sys.stdout.flush()

        self.shap_plots(xgb_clf_no_nor, all_num, new_prefix, _groups_dict)
        joblib.dump(xgb_clf_no_nor, model_filename)

        self.create_metdata(
            new_prefix = new_prefix,
            values = [
                model_filename,
                accuracy,
                _groups_dict[True],
                _groups_dict[False]
            ]
        )

        return (all_num, all_labels, 
                _groups_dict, xgb_clf_no_nor, 
                dataset)

    def confusion_plots(self, mytables):

        if not mytables:
            return None

        cnf_mx_filename = "cnf_mx_%s.png" % self.model_prefix

        if len(self.tree_id_comp)  <= 1:

            X, y,_groups_dict,xgb_clf_no_nor,title = mytables[0]
            plot_confusion_matrix(
                xgb_clf_no_nor, X, y,
                values_format  = 'd',
                display_labels = [_groups_dict[i] for i in xgb_clf_no_nor.classes_]
            )
            # axes[i].set_title(title, fontsize = 20)
            plt.savefig(cnf_mx_filename, bbox_inches = 'tight')
            plt.close()
            

        else: 
            if len(self.tree_id_comp) <= 3:
                nrows = 1
                self.cnf_ncols = len(self.tree_id_comp)
                
            else:
                res = len(mytables) % self.cnf_ncols
                nrows = (len(mytables) // self.cnf_ncols) + bool(res)

            f,axes = plt.subplots(nrows = nrows, ncols = self.cnf_ncols, figsize=(18, 5), dpi = 400)
            for i in range(len(mytables)):
                X, y,_groups_dict,xgb_clf_no_nor,title = mytables[i]

                plot_confusion_matrix(
                    xgb_clf_no_nor, X, y,
                    values_format  = 'd',
                    display_labels = [_groups_dict[i] for i in xgb_clf_no_nor.classes_],
                    ax = axes[i]
                )
                axes[i].set_title(title, fontsize = 20)

            plt.savefig(cnf_mx_filename, bbox_inches = 'tight')
            plt.close()

    def xgboost_iterator(self):

        mytables = []
        for tree_id_comp1 in self.tree_id_comp:

            tmp_model = self.xgboost_classifier(tree_id_comp1)
            if tmp_model:
                mytables.append(tmp_model)

        sys.stdout.write("Plotting confusion matrices\n")
        sys.stdout.flush()

        self.confusion_plots(mytables)

# debugging --------------------------------------

# import os
# base_path = '/Users/ulises/Desktop/GOL/software/GGpy/proofs_ggi/flatfishes'

# file_comparisons = os.path.join(base_path, 'comparisons.txt')
# # features_file    = '/Users/ulises/Desktop/ABL/GGI_flatfishes/post_ggi/features_fstats_yongxin.tsv'
# features_file    = os.path.join(base_path, 'features_991exons_fishlife.tsv')
# all_ggi_results  = os.path.join(base_path, 'ggi_partial_results_clean.tsv')
# threads = 5

# self = Post_ggi(
#     feature_file = features_file,
#     all_ggi_results = all_ggi_results,
#     file_comparisons = file_comparisons,
#     threads = 6
# )

# self.xgboost_classifier_iterator()

# debugging --------------------------------------




import os
import re
import sys
import csv
import ggpy
import pickle
import dendropy


from multiprocessing import Pool
from ggpy.utils import fas_to_dic, remove_files
from ggpy.wrappers import Raxml, RAXML, Consel


class GGI(Raxml, Consel):

    def __init__(self, 
                sequences = None, 
                taxonomyfile = None,
                topologies = None,
                are_extended = False,
                rooted = False,
                codon_partition = True, 
                threads = 1,
                parallel_gt = True,

                evomodel = 'GTRGAMMA',
                iterations = 10,

                write_extended = False,
                suffix = 'ggi.tsv',

                raxml_exe = RAXML, #unmutable vars

                ) -> None:


        self.sequences = sequences
        self.taxonomyfile = taxonomyfile

        self.rooted = rooted

        self.topologies = topologies
        self.are_extended = are_extended
        self.write_extended = write_extended


        self.suffix = suffix
        


        ## --- shared variables --- #
        self.threads = threads
        self.parallel_gt = parallel_gt


        self.codon_partition = codon_partition
        self.evomodel = evomodel
        self.raxml_exe = raxml_exe
        self.iterations = iterations

        self.seqmt_exe   = "seqmt"
        self.makermt_exe = "makermt"
        self.consel_exe  = "consel"
        self.catpv_exe   = "catpv"
        ## --- shared variables --- #

        self.translation = {}
        self.allheaders = set()
        self.final_out = "out_" + self.suffix
        self.trans_out = "translation_" + self.suffix

        self.unable_to_prune_f = "unable_to_prune_" + self.suffix
        self.unable_to_run_f = "unable_to_run_" + self.suffix

        self.hiddenfile = ".treebase_" + self.suffix

    
    @property
    def _taxa(self):

        if not self.taxonomyfile:
            return None

        myrows = []
        with open(self.taxonomyfile, 'r') as f:
            reader = csv.reader(f)
            for row in reader:
                myrows.append(row)
                
        df = {}
        for i in myrows:
            spps  = i[0]
            group = i[1]
            if not df.__contains__(group):
                df[group] = [spps]
            else:
                df[group] += [spps]

        return df

    def _internal_hypothesis(self, df, ext_str):

        ngroups = len(df)

        file_basename = "g_%s_%s" % ( ngroups, ext_str         )
        file_path     = os.path.join( ggpy.__path__[0], 'data' )
        filename      = os.path.join( file_path, file_basename )

        fmt_args_exte = {}
        fmt_args_cons = {}

        g_count  = 1
        for k,v in df.items():

            quoted_spps = ",".join(v)
            if len(v) > 1:
                quoted_spps = "(%s)" % quoted_spps

            fmt_args_exte[ "g_%s" % g_count ] = quoted_spps
            fmt_args_cons[ "g_%s" % g_count ] = k
            g_count += 1

        with open(filename, 'r') as f:
            tmp_str = f.read()

        filled_spps_str  = tmp_str.format(**fmt_args_exte).split()
        filled_group_str = tmp_str.format(**fmt_args_cons).split()

        self.translation = dict(zip(filled_spps_str, filled_group_str))

        hypothesis = {}
        count = 1
        for spps,group in zip(filled_spps_str, filled_group_str):
            hypothesis[count] = {'group': group, 'extended': spps}
            count += 1

        self.translation = hypothesis

    def _taxa_from_str(self, tmp_tree):

        tmp_tree = dendropy.Tree.get_from_string(  
                                    src = tmp_tree, 
                                    schema = 'newick',
                                    preserve_underscores = True
                                )

        return [ i.taxon.label for i in tmp_tree.leaf_node_iter() ]

    def _external_hypothesis(self, df):

        groups = set(df)
        hypothesis = {}

        with open(self.topologies, 'r') as f:

            count = 1
            for i in f.readlines():

                tmp_hypothesis = i.strip()

                if not tmp_hypothesis:
                    continue

                str_taxa = self._taxa_from_str(tmp_hypothesis)

                new_groups = set(str_taxa) - groups

                if new_groups:
                    sys.stderr.write( "\n %s groups do not match with the taxonomy file at: '%s'\n" % (new_groups, tmp_hypothesis) )
                    sys.stderr.flush()
                    continue
      
                extended_str = tmp_hypothesis
                for k,v in df.items():

                    quoted_spps = ",".join(v)
                    if len(v) > 1:
                        quoted_spps = "(%s)" % quoted_spps

                    extended_str = extended_str.replace(k, quoted_spps)

                hypothesis[count] = {'group':tmp_hypothesis, 'extended': extended_str}
                count += 1

        if not hypothesis:
            # no hypothesis
            # to test
            self.close_files()
            sys.exit(1)

        self.translation = hypothesis

    def __save_trees__(self, obj = None):

        with open( self.hiddenfile , 'wb') as f:
            pickle.dump(obj, f, pickle.DEFAULT_PROTOCOL)

    def __load_trees__(self):

        with open( self.hiddenfile, 'rb') as f:
            return pickle.load(f)

    def _set_extended_hypothesis(self):

        if self.topologies and self.are_extended:

            to_trs = {}
            with open(self.topologies, 'r') as f:

                count = 1 
                for i in f.readlines():
                    tmp_hypothesis = i.strip()
                    to_trs[count] = {'group': None, 'extended': tmp_hypothesis}
                    count += 1

            self.translation = to_trs

        else:

            ngroups = len(self._taxa)

            if not self.topologies:

                ext_str = ""

                if ngroups == 3:
                    self.rooted = True

                if self.rooted:
                    if ngroups > 8:
                        sys.stderr.write("\nTest all possible rooted trees for more \n")
                        sys.stderr.write("than 8 groups is not currently available\n")
                        sys.stderr.flush()
                        self.close_files()
                        sys.exit(1)

                    ext_str += "rooted.tree"
                else:

                    if ngroups > 9:
                        sys.stderr.write("\nTest all possible unrooted trees for more \n")
                        sys.stderr.write("than 9 groups is not currently available\n")
                        sys.stderr.flush()
                        self.close_files()
                        sys.exit(1)

                    ext_str += "unrooted.tree"

                self._internal_hypothesis(self._taxa, ext_str)
            else:

                self._external_hypothesis(self._taxa)

        if self.write_extended:

            outname = "extended_hypothesis_" + self.suffix
            with open(outname, "w") as f:
                for v in self.translation.values():
                    f.write("%s\t%s\n" % (v['group'], v['extended']))

            sys.stderr.write("\nExtended hypothesis written at '%s'\n" % outname)
            sys.stderr.flush()
            self.close_files()
            sys.exit(0)

        # attempt to reduce mem usage
        self.__save_trees__(self.translation)
        self.translation = {}
        
    def _aln_in_hypothesis(self, aln_taxa, tmp_tree):
        """
        # hypothesis taxa includes all aln taxa?
        """
        # str_hypothesis = list(self.translation)[0]
        # hypo_taxa = set(self._taxa_from_str(str_hypothesis))

        hypo_taxa = set( [i.taxon.label for i in tmp_tree.leaf_node_iter()] )
        aln_taxa  = set( aln_taxa )

        new_for_hypo = aln_taxa - hypo_taxa

        return (False, new_for_hypo) if new_for_hypo else (True, None)

    def prunning(self, file):
        # file = self.sequences[0]
        aln_base = os.path.basename(file)
        aln      = fas_to_dic(file)

        translation  = self.__load_trees__()
        aln_metadata = {}

        for k,v in translation.items():
            # k,v = tuple(self.translation.items())[0]

            tmp_tree = (dendropy
                            .Tree
                            .get_from_string( 
                                src = v['extended'], 
                                schema = 'newick',
                                preserve_underscores = True) )

            aln_taxa  = [i.replace(">", "") for i in list(aln)]
            is_subset,which_taxa = self._aln_in_hypothesis(aln_taxa, tmp_tree)

            if not is_subset:
                tmp_rows = [ [aln_base, wt, v['id']] for wt in which_taxa ]
                with open(self.unable_to_prune_f, 'a') as f:
                    writer = csv.writer(f, delimiter = "\t")
                    writer.writerows(tmp_rows)

                continue

            tmp_tree.retain_taxa_with_labels( aln_taxa )
            pruned = tmp_tree.as_string(schema='newick')

            aln_metadata[k] = {
                    'group' : v['group'], 
                    'pruned': pruned , 
                    'aln'   : file
                }

        del translation
        return aln_metadata

    def constraints(self, pr_message):
        """
        What it bassically runs:

        raxmlHPC-SSE3_{OS}_64bit                             \\
            -p 12345                                         \\
            -g {fasta}_{id}_{suffix}_constr.tree             \\
            -q {fasta}_{id}_{suffix}.partitions #if selected \\
            -m {model}                                       \\
            -s {fasta}_{id}_{suffix}                         \\
            -n base({fasta}_{id}_{suffix}) + ".tree"         \\
            -T 1                                             \\
            --silent                                         \\
            -N {iterations}
        """
        if not pr_message:
            return None

        return self.__iter_raxml__(pr_message, 
                                   self.unable_to_run_f, 
                                   self.suffix)
        
    def site_likelihoods(self, cons_message, aln_f):
        """
        constraints are sent 
        ordered in functions of keys\\

        keys are ordered in functions
        of row lines in the
        translation file
        """
        # cons_message, aln_f = cons_message, file 
        if not cons_message:
            return None

        # aln_f = next(iter(cons_message.values()))['aln']
        whole_constrs_f = aln_f + "_AllCons_" + self.suffix
        
        # join constrains
        cons_trees = []
        # order matters
        for i in range(len(cons_message)):
            v = cons_message[ i + 1 ]
            with open( v['constrained'], 'r' ) as c:
                cons_trees.append( c.read() )

        with open( whole_constrs_f, "w") as f:
            f.writelines(cons_trees)

        sll_f = self._site_likehood(
                        (aln_f, whole_constrs_f),
                        self.suffix,
                        self.unable_to_run_f
                )

        if sll_f:
            return (sll_f, cons_message)
            # return (sll_f, sll_meta)
        else:
            return None

    def keep_rank1_tree(self, au_table, sll_meta):
        to_rm = []
        s_item = 0

        for rank, item, _ in au_table:
            item_int = int(item)

            if rank != '1':
                to_rm.append(

                    sll_meta[item_int]['constrained']
                )

            else:
                s_item += item_int
                # rank1 += sll_meta[item_int]['constrained']

        rank1 = sll_meta[s_item]['constrained']
        # seq = sll_meta[s_item]['aln']
        # self.add_bootstrap(
        #     seq = seq, 
        #     selected_tree = rank1,
        #     suffix = self.suffix
        # )
        new_name = re.sub(self.suffix, self.suffix + "_rank1", rank1)

        os.rename(rank1, new_name)
        remove_files(to_rm)

    def au_tests(self, sll_message, seq_basename):

        if not sll_message:
            return None

        sll_f,sll_meta = sll_message

        au_table = self._consel_pipe(seq_basename, sll_f, self.unable_to_run_f)

        if not au_table:
            return None

        self.keep_rank1_tree(au_table, sll_meta)

        out_cols = []
        for row in au_table:
            item    = int(row[1])

            rank    = row[0]
            au_test = row[2]
            tree_id = item
            group   = sll_meta[item]['group']

            out_cols.append([
                    seq_basename, tree_id,
                    group, rank, au_test
                ])

        return out_cols

    def ggi_iterator(self, file):
        # file = self.sequences[0]
        seq_basename = os.path.basename(file)

        sys.stdout.write("Processing: '%s' \n" % seq_basename)
        sys.stdout.flush()

        pr_message   = self.prunning(file)
        cons_message = self.constraints(pr_message)
        sll_message  = self.site_likelihoods(cons_message, file)
        
        return self.au_tests(sll_message, seq_basename)

    def init_files(self):

        utp_cols = [["alignment", "species", "tree_id"]]
        with open(self.unable_to_prune_f, "w") as f:
            writer = csv.writer(f, delimiter = "\t")
            writer.writerows(utp_cols)

        utr_cols = [["alignment", "where", "tree_id"]]
        with open(self.unable_to_run_f, "w") as f:
            writer = csv.writer(f, delimiter = "\t")
            writer.writerows(utr_cols)

    def close_files(self):
        files_check = [
            self.unable_to_prune_f,
            self.unable_to_run_f
        ]

        to_rm = [self.hiddenfile]

        sys.stdout.write("\n\n")
        sys.stdout.flush()

        for to_c in files_check:
            num_lines = sum( 1 for _ in open(to_c) )
            if num_lines == 1:
                to_rm.append(to_c)
            else:
                sys.stdout.write("'%s' file was written\n" % to_c)
                sys.stdout.flush()
                
        remove_files(to_rm)

    def export_translation(self):
                
        trans_cols = [[ "tree_id", "group", "extended"]]
        translation = self.__load_info__(self.hiddenfile)

        trans_sorted = sorted(translation.items(), key = lambda kv: kv[0])

        for k,v in trans_sorted:
            trans_cols.append([ k, v['group'], v['extended'] ])

        with open(self.trans_out, "w") as f:
            writer = csv.writer(f, delimiter = "\t")
            writer.writerows(trans_cols)

        sys.stdout.write("=> '%s' file was written\n" % self.trans_out)
        sys.stdout.flush()

    def oneCore_gt(self):

        preout = []
        with Pool(processes= self.threads) as p:

            for fa in self.sequences:
                result  = p.map_async(self.ggi_iterator, (fa,))
                preout.append(result)

            out_cols = [["alignment", "tree_id", "group", "rank", "au_test"]]
            for pr in preout:
                gotit = pr.get()[0]
                if gotit:
                    out_cols.extend(gotit)

        return out_cols        
    
    def manyCore_gt(self):

        # preout = []
        out_cols = [["alignment", "tree_id", "group", "rank", "au_test"]]
        for fa in self.sequences:
            result = self.ggi_iterator(fa)
            if result:
                out_cols += result

        return out_cols

    def main(self):

        self.init_files()
        self._set_extended_hypothesis()

        if not self.parallel_gt:
            out_cols = self.oneCore_gt()

        else:
            out_cols = self.manyCore_gt()


        if len(out_cols ) > 1:    
            sys.stdout.write("\n\n")
            sys.stdout.flush()

            with open(self.final_out, "w") as f:
                writer = csv.writer(f, delimiter = "\t")
                writer.writerows(out_cols)

            sys.stdout.write("=> '%s' file was written\n" % self.final_out)
            sys.stdout.flush()

            self.export_translation()
            self.close_files()
        else:

            self.close_files()
            sys.exit(1)
                

# tests ----------------------#
# import glob
# # sequences = glob.glob("/Users/ulises/Desktop/GOL/software/GGpy/demo/LOC*.fas")
# sequences = glob.glob("/Users/ulises/Desktop/GOL/software/GGpy/demo/E*.fasta")


# self = GGI(
#     sequences=sequences,
#     # taxonomyfile=None,
#     taxonomyfile="/Users/ulises/Desktop/GOL/software/GGpy/demo/ggi_tax_file.csv",
#     # topologies= "/Users/ulises/Desktop/BAL/GGI_flatfishes/tests/all_constraints_hypos.trees",
#     topologies= "/Users/ulises/Desktop/GOL/software/GGpy/demo/myhypothesis.trees",
#     write_extended= False,
#     # are_extended = False,
#     are_extended = False,
#     codon_partition=False,
#     iterations=1,
#     threads= 2,
#     parallel_gt=True
#     )

# self._set_extended_hypothesis()
# tests ----------------------#

class deepGGI:
    def __init__(self) -> None:
        pass



import os
import re
import sys
import operator
import subprocess



def runshell(args, type = ""):
    
    if type == "stdout":
        a, b = args
        with open(b, "w") as f:
            subprocess.call(a, stdout= f)

    else:
        p = subprocess.Popen(args)
        p.communicate()


def export_fasta(aln: dict, outname: str):
    """
    This function assumes sequences names
    are already formated (i.e., names starting with ">")
    """
    with open(outname, 'w') as f:
        for k,v in aln.items():
            f.write( "%s\n%s\n" % (k,v))


def fas_to_dic(file):
    
    file_content = open(file, 'r').readlines()
    seqs_list   = []
    
    for i in file_content:
        line = i.strip()
        if line: # just if file starts empty
            seqs_list.append(line) 
    
    keys = [] 
    values = []    
    i = 0
    while(">" in seqs_list[i]):
        keys.append(seqs_list[i])
        i += 1 
        JustOneValue = []

        while((">" in seqs_list[i]) == False):
            JustOneValue.append(seqs_list[i]) 
            i += 1

            if(i == len(seqs_list)):
                i -= 1
                break

        values.append("".join(JustOneValue).upper().replace(" ", ""))
        
    return dict(zip(keys, values))


def remove_files(files):

    for f in files:
        try:
            os.remove(f)
        except FileNotFoundError:
            pass

def check_empty_positions(aln, file, outname):
    # file = '/Users/ulises/Desktop/GOL/software/GGpy/demo/toggi/E1644.fasta_2_ggi.tsv'
    # aln = fas_to_dic(file)
    gap_chars = set(['N', '-', '!', '?'])
    seq_len = len(next(iter(aln.values())))

    all_char_pos1 = set()
    all_char_pos2 = set()
    all_char_pos3 = set()

    for v in aln.values():
        for p in range(0, seq_len, 3):
            all_char_pos1.update(v[p])
            all_char_pos2.update(v[p + 1])
            all_char_pos3.update(v[p + 2])

    is_pos1_empty = not (all_char_pos1 - gap_chars)
    is_pos2_empty = not (all_char_pos2 - gap_chars)
    is_pos3_empty = not (all_char_pos3 - gap_chars)

    all_empty = is_pos1_empty and is_pos2_empty and is_pos3_empty
    all_filled = not is_pos1_empty and not is_pos2_empty and not is_pos3_empty

    if all_empty:
        sys.stderr.write("'%s' file has only gap characters\n" % file)
        sys.stderr.flush()
        sys.exit(1)

    if all_filled:
        return False

    new_aln = {}
    for k,v in aln.items():
        mystr = ""
        for i in range(0, seq_len, 3):

            F = '' if is_pos1_empty else v[i]
            S = '' if is_pos2_empty else v[i + 1]
            T = '' if is_pos3_empty else v[i + 2]
            mystr += (F + S + T)

        new_aln[k] = mystr
    
    seq_new_len = len(next(iter(new_aln.values())))

    filled = 3 - sum([is_pos1_empty, is_pos2_empty, is_pos3_empty])
    with open(outname, 'w') as outf:
        for i in range(filled):
            outf.write(
                "DNA, p{pos} = {pos}-{seq_len}\\{npart}\n".format(
                    pos = i + 1,
                    seq_len = seq_new_len,
                    npart = filled
                )
            )
    
    with open(file, 'w') as f:
        for k,v in new_aln.items():
            f.write( "%s\n%s\n" % (k,v))
            
    return True

def codon_partitions(file, outname = None):

    aln = fas_to_dic(file)
    lengths = set([len(v) for _,v in aln.items()])

    if lengths.__len__() > 1:
        sys.stderr.write("'%s' file has sequences with different lengths\n" % file)
        sys.stderr.flush()
        sys.exit(1)

    procesed = check_empty_positions(aln, file, outname)

    if procesed:
        return None
    
    aln_length = lengths.pop()
    
    with open(outname, 'w') as outf:
        outf.write("DNA, p1 = 1-%s\\3\n" % aln_length)
        outf.write("DNA, p2 = 2-%s\\3\n" % aln_length)
        outf.write("DNA, p3 = 3-%s\\3\n" % aln_length)

    return None

def _filter(indexes: list, aln: dict, seqlength: int, offset: int) -> dict:
    """
    if indexes list is empty, it means
    there is anything to save because any column
    pass the maximun allowed proportion condition.

    then, return None
    """

    if not indexes:
        return None

    out = {}
    for k,v in aln.items():
        mystr  = ""
        for c in range(0, seqlength, offset):
            if c in indexes:
                mystr += v[c:c+offset]
        out[k] = mystr

    seqlength = _seq_length(out)

    if not seqlength:
        return None

    return out

def _get_gap_prop(aln: dict, seqlength: int, offset: int)-> dict:

    idxs = []
    for c in range(0, seqlength, offset):

        mystr = ""
        for v in aln.values():
            mystr += v[c:c+offset]

        gap_prop = mystr.count('-')/len(mystr)
        idxs.append( (c, gap_prop) )

    return idxs

def _seq_length(aln:dict)-> int:
    return len(next(iter(aln.values())))

def convertMissing2Gap(aln: dict, isprot = False) -> dict:

    for k,v in aln.items():
        if isprot:
            aln[k] = v.replace("X", "-")

        else:
            aln[k] = v.replace("N", "-")

    return aln

def close_gaps(aln: dict, is_codon_aware: bool = True, model: str = 'GTRGAMM') -> dict:
    """
    Moves:
    1. It converts N to gaps
    2. get proportions
    3. filter


    * If `aln` is empty, returns None
    * If `aln` is not empty, `aln` has values with length
    * If not seqlength after processing, None is returned
    """
    # aln = {'a':'---cat--a', 'b':'---cat--a'}
    # is_codon_aware = not True
    if not aln:
        return None
        
    offset    = 3 if is_codon_aware else 1
    aln       = convertMissing2Gap( aln = aln, isprot = re.findall('^PROT', model) )
    seqlength = _seq_length(aln)

    idxs = []
    idxs_prop = _get_gap_prop(aln, seqlength, offset)

    for c, gap_prop in idxs_prop:
        if not gap_prop == 1:
            idxs.append(c)

    return _filter(idxs, aln, seqlength, offset)



#!/usr/bin/env python

import sys
import platform

import setuptools
from distutils.core import setup

if platform.architecture()[0] != '64bit':
    sys.stderr.write('Architecture requires 64bit')
    sys.stderr.flush()
    exit()

myos = sys.platform

if myos == 'darwin':
    bins = [
        # './ext_bin/raxml/raxmlHPC-SSE3_Darwin_64bit',
        './ext_bin/raxml/raxmlHPC-PTHREADS-SSE3_Darwin_64bit',
        './ext_bin/consel/darwin/seqmt',
        './ext_bin/consel/darwin/makermt',
        './ext_bin/consel/darwin/consel',
        './ext_bin/consel/darwin/catpv'
            ]

elif myos == 'linux' or myos == "linux2":
    bins = [
        # './ext_bin/raxml/raxmlHPC-SSE3_Linux_64bit',
        './ext_bin/raxml/raxmlHPC-PTHREADS-SSE3_Linux_64bit',
        './ext_bin/consel/linux/seqmt',
        './ext_bin/consel/linux/makermt',
        './ext_bin/consel/linux/consel',
        './ext_bin/consel/linux/catpv'
            ]
else:
    sys.stderr.write('Package does not work with %s operative system'  % myos)
    sys.stderr.flush()
    exit()


dependencies = [
    "fishlifetraits>=0.4.5", # own package
    'dendropy==4.4.0',
    'shap',
    'joblib',
    'xgboost',
    'pandas',
    'scikit-learn',
    'matplotlib',
]

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(name = "ggi",
      version = '1.2.0',
      maintainer = 'Ulises Rosas',
    #   long_description = readme,
    #   long_description_content_type = 'text/markdown',
      url ='https://github.com/Ulises-Rosas/GGpy',
      packages = ['ggpy'],
      package_data = {'ggpy': ['data/*']} ,
      data_files = [ ('bin', bins) ],
      include_package_data=True,
      install_requires = dependencies,
      zip_safe = False,
      entry_points = {
        'console_scripts': [
            'ggpy   = ggpy.cli:main'
            ]
      },
      scripts=[
          './scripts/root_groups.py',
      ],
      classifiers = [
          'Programming Language :: Python :: 3'
      ]
    )
